#!/bin/bash
#
# Script to automatically scan a page to a file.
#
# Version 1.2
#
# Copyright (C) 2009 Helmar Gerloni
#
# Author: Helmar Gerloni <helmar@gerloni.net>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

set -e

TMPO=$(tempfile)

trap "echo" INT TERM ERR
trap "[ $TMPO ] && rm -f $TMPO $TMPO.txt $TMPO.pnm $TMPO.tif $TMPO.pdf $TMPO.cat.pdf $TMPO.opt.pdf $TMPO.ps" EXIT

COMMAND=`basename $0`

declare -i RESOL

echo

################################################################################
# Default settings
################################################################################

# (*) Some of these options depend on your scanner. See
# 'scanimage --help --device DEVICE' for supported options and values.

# Input device name (scanner) to use: If you have more than one scanner
# connected to your system then you can specify the one you want to use here.
# See 'scanimage --list-devices', e.g.
# SCANNER="Plustek OpticSlim 2400 flatbed scanner"
SCANNER=""

# Scan mode: lineart (black/white), gray (grayscale), or color (*)
MODE=lineart

# Resolution: 150|300|600|1200 dpi (*)
RESOL=300

# Page size: a4|a5|letter
PAGESIZE="a4"
# Page settings. (*)
# See 'man scanimage' for details.
PAGESETTINGS_A4="-l 2 -t 2 -x 210 -y 297"
PAGESETTINGS_A5="-l 2 -t 2 -x 148 -y 210"
PAGESETTINGS_LETTER="-l 0 -t 2 -x 216 -y 279"

# Metadata for tif or pdf files
AUTHOR=""
TITLE=""
SUBJECT=""
KEYWORDS=""

# Print usage
################################################################################
function usage {
  echo "Usage: $COMMAND [OPTIONS]"
  echo
  echo "Scan a page and write it to a pnm, tif or pdf file."
  echo "The main purpose of this script is to digitally archive documents."
  echo
  echo "-d SCANNER"
  echo "  Specifies the device or name of the scanner you want to use."
  echo "  Use 'scanimage --list-devices' to show all available scanners."
  echo "  It is also possible to specify a pnm file for SCANNER. Then instead of doing"
  echo "  a scan the given file is used as input."
  echo "  Default:"
  echo "    If no scanner is given the first available scanner is used."
  echo "  i.e. '-d \"gt68xx:libusb:002:002\"'"
  echo "       '-d \"Plustek OpticSlim 2400 flatbed scanner\"'"
  echo "       '-d \"/path/to/image.pnm\"'"
  echo
  echo "-m MODE"
  echo "  Selects the scan mode, usually lineart|gray|color, but depends on your"
  echo "  scanner. See 'scanimage --help --device DEVICE' for supported values."
  echo "   Default mode: lineart (black/white)."
  echo "   i.e. '-m gray'"
  echo
  echo "-r RESOLUTION"
  echo "  Sets the resolution of the scanned image."
  echo "  Typical values are 150|300|600|1200 dpi, but your scanner may use others."
  echo "  See  'scanimage --help --device DEVICE' for supported values."
  echo "  Default resolution: 300."
  echo "  i.e. '-r 600'"
  echo
  echo "-f txt|pnm|tif|pdf"
  echo "  Format of the output file. If the output format is txt then optical"
  echo "  character recognition (OCR) is done."
  echo "  Default:"
  echo "    If '-o OUTFILE.xyz' is given the format xyz is used."
  echo "    If both -f and -o are not given the pdf format is used."
  echo
  echo "-p a4|a5|letter"
  echo "  Selects the page size."
  echo "  Default size: a4."
  echo
  echo "-o OUTFILE.xyz"
  echo "-o print:[PRINTER]"
  echo "  Write the output to the specified file, directory or printer."
  echo "  Use 'lpstat -a' to show all available printers."
  echo "  When using the tif or pdf format the new scan is appended"
  echo "  to OUTFILE.xyz if the given file already exists."
  echo "  If the given output file already exists the old file is moved"
  echo "  to OldScan0001-OUTFILE.xyz, OldScan0002-OUTFILE.xyz, ..."
  echo "  i.e. '-o \"path/to/output.pdf\"'"
  echo "       '-o \"path/to/outdir/\"'"
  echo "       '-o print:'"
  echo "       '-o print:Cups_Printer_Name'"
  echo "  Default:"
  echo "    If no output filename is given or if the given name is a"
  echo "    directory the output filename defaults to"
  echo "    Scan0001.format, Scan0002.format, ..."
  echo "    If no printer is given the default printer shown by 'lpstat -d' is used."
  echo
  echo "-a AUTHOR"
  echo "-t TITLE"
  echo "-s SUBJECT"
  echo "-k KEYWORDS"
  echo "  Sets the document information author/title/subject/keywords for tif and pdf"
  echo "  files."
  echo
  echo "-h"
  echo "  Display this help message and exit."
  echo
}

# Parse command line options
################################################################################
while getopts "d:m:r:f:p:o:a:t:s:k:h" CMDopt
do
  case $CMDopt in
    d) SCANNER="$OPTARG";;
    m) MODE="$OPTARG";;
    r) RESOL="$OPTARG";;
    f) OUTFORMAT="$OPTARG";;
    p) PAGESIZE="$OPTARG";;
    o) OUTFILE="$OPTARG";;
    a) AUTHOR="$OPTARG";;
    t) TITLE="$OPTARG";;
    s) SUBJECT="$OPTARG";;
    k) KEYWORDS="$OPTARG";;
    *) usage
       [ $CMDopt == "h" ] && exit || exit 1
       ;;
  esac
done
# No argument should be left here
shift $((OPTIND-1))
if [ $1 ]; then
  echo "Unknown argument $1."
  usage
  exit 1
fi

# Check and evaluate command line options
################################################################################
if [ `echo $OUTFILE | grep ^print:` ]; then
  # Output to printer
  PRINTER=`echo $OUTFILE | sed 's/^print://'`
  OUTFILE="print:"
else
  # Output to file
  # Remove unnecessary slashes from OUTFILE
  OUTFILE=`echo "$OUTFILE" | sed 's/\/\+/\//g'`
  [ "$OUTFILE" == "/" ] || OUTFILE=`echo "$OUTFILE" | sed 's/\/$//g'`
  # If OUTFORMAT is not given try to determine it from OUTFILE suffix
  if [ -z "$OUTFORMAT" ]; then
    OUTFORMAT=`echo "$OUTFILE" | grep "\." | awk -F . '{ print $NF }'`
  fi
  # Convert OUTFORMAT to all lowercase letters
  OUTFORMAT=`echo "$OUTFORMAT" | sed  "s/./\L&/g"`
  # OUTFORMAT defaults to pdf
  if [ -z "$OUTFORMAT" ]; then
    OUTFORMAT="pdf"
  elif [ "$OUTFORMAT" != "txt" ] && [ "$OUTFORMAT" != "pnm" ] && [ "$OUTFORMAT" != "tif" ] && [ "$OUTFORMAT" != "tiff" ] && [ "$OUTFORMAT" != "pdf" ]; then
    echo "Warning: Unknown output format '$OUTFORMAT', using 'pdf' instead." >&2
    OUTFORMAT="pdf";
  fi
  # Split OUTFILE in [OUTDIR/]OUTFILE
  if [ "$OUTFILE" ]; then
    if [ -d "$OUTFILE" ]; then
      OUTDIR="$OUTFILE/"
      OUTFILE=""
    else
      OUTDIR=`dirname "$OUTFILE"`/
      OUTFILE=`basename "$OUTFILE"`
      if [ ! -e "$OUTDIR" ]; then
        echo "Error: Output directory '$OUTDIR' not found." >&2
        echo
        exit 1
      fi
    fi
  fi
  [ "$OUTDIR" == "./" ] && OUTDIR=""
  [ "$OUTDIR" == "//" ] && OUTDIR="/"
  # Find a valid output filename
  if [ -z "$OUTFILE" ]; then
    SCANNR="1"
    PAD="000"
    while [ -e "$OUTDIR""Scan$PAD$SCANNR.$OUTFORMAT" ]; do
      SCANNR=$(($SCANNR + 1))
      PAD=""
      [ "$SCANNR" -lt 1000 ] && PAD="0"
      [ "$SCANNR" -lt 100  ] && PAD="00"
      [ "$SCANNR" -lt 10   ] && PAD="000"
    done
    OUTFILE="Scan$PAD$SCANNR.$OUTFORMAT"
  elif [ -e "$OUTDIR$OUTFILE" ]; then
    SCANNR="1"
    PAD="000"
    while [ -e "$OUTDIR""OldScan$PAD$SCANNR-$OUTFILE" ]; do
      SCANNR=$(($SCANNR + 1))
      PAD=""
      [ "$SCANNR" -lt 1000 ] && PAD="0"
      [ "$SCANNR" -lt 100  ] && PAD="00"
      [ "$SCANNR" -lt 10   ] && PAD="000"
    done
    OUTFILE_OLD="OldScan$PAD$SCANNR-$OUTFILE"
    echo "Output file '$OUTDIR$OUTFILE' already exists. Moving it to '$OUTDIR$OUTFILE_OLD'."
    mv "$OUTDIR$OUTFILE" "$OUTDIR$OUTFILE_OLD"
  fi

  [ "$OUTFORMAT" == "tiff" ] && OUTFORMAT="tif"

fi
# Convert MODE/PAGESIZE to all lowercase letters
MODE=`echo "$MODE" | sed  "s/./\L&/g"`
PAGESIZE=`echo "$PAGESIZE" | sed  "s/./\L&/g"`
# Check given PAGESIZE
case "$PAGESIZE" in
  a4)     PAGESETTINGS="$PAGESETTINGS_A4";;
  a5)     PAGESETTINGS="$PAGESETTINGS_A5";;
  letter) PAGESETTINGS="$PAGESETTINGS_LETTER";;
  *)      echo "Warning: Unknown page size '$PAGESIZE', using 'a4' instead." >&2
          PAGESIZE="a4"
          PAGESETTINGS="$PAGESETTINGS_A4"
          ;;
esac
# Check given RESOL
if [ "$RESOL" -lt 1 ]; then
  echo "Warning: Invalid resolution $RESOL dpi." >&2
  echo "         Using 300 dpi instead." >&2
  RESOL=300
fi

# Get the source image
################################################################################
if [ -f "$SCANNER" ]; then
  # No scanning needed, go on with the given file
  echo "Not really scanning, using file '$SCANNER' as input."
  echo
  cp "$SCANNER" "$TMPO.pnm"
else
  # Find a scanner
  echo -n "Searching scanner... "
  SCANNERS=`scanimage --formatted-device-list '^%d^ ^%v %m %t^\' | sed 's/\\\/\\n/g'`
  echo  "done."
  if [ -z "$SCANNERS" ]; then
    echo "Error: No scanner found on your system." >&2
    echo
    exit 1
  fi
  if [ "$SCANNER" ]; then
    SCANDEV=`echo "$SCANNERS" | grep "\^$SCANNER\^" | awk -F ^ '{ print $2 }'`
    if [ -z "$SCANDEV" ]; then
      echo "Warning: File or scanner '$SCANNER' not found." >&2
    fi
    SCANNER=`echo "$SCANNERS" | grep "\^$SCANNER\^" | awk -F ^ '{ print $4 }'`
  fi
  if [ -z "$SCANDEV" ] && [ `echo "$SCANNERS" | wc -l` -gt 1 ]; then
    echo "Warning: More than one scanner found:" >&2
    echo "$SCANNERS" | sed "s/^\^/         - Device '/" | sed "s/\^ \^/' is a '/" | sed "s/\^$/'/" >&2
    echo "         Use the -d option to specify the scanner you want to use." >&2
  fi
  if [ -z "$SCANDEV" ]; then
    SCANDEV=`echo "$SCANNERS" | head -1 | awk -F ^ '{ print $2 }'`
    SCANNER=`echo "$SCANNERS" | head -1 | awk -F ^ '{ print $4 }'`
  fi
  # Scan
  echo "Using the '$SCANNER' (device '$SCANDEV') with settings:"
  echo "  Scan mode:        $MODE"
  echo "  Resolution:       $RESOL dpi"
  echo "  Page size:        $PAGESIZE"
  scanimage $PAGESETTINGS --mode "$MODE" --resolution "$RESOL" --device-name "$SCANDEV" --progress > "$TMPO"
  echo
  pnmflip -rotate180 "$TMPO" > "$TMPO.pnm"

fi

# Generate output
################################################################################
if [ "$OUTFILE" == "print:" ]; then
  pnmtotiff "$TMPO.pnm" > "$TMPO.tif"
  tiff2pdf -r d -x "$RESOL" -y "$RESOL" -j -p "$PAGESIZE" -o "$TMPO.pdf" "$TMPO.tif"
  pdftops "$TMPO.pdf" "$TMPO.ps"
  if [ "$PRINTER" ]; then
    echo "Sending to printer '$PRINTER'... "
    lp -t "Scan from $SCANNER" -d "$PRINTER" "$TMPO.ps"
  else
    echo "Sending to default printer (`lpstat -d | awk '{ print $NF }'`)... "
    lp -t "Scan from $SCANNER" "$TMPO.ps"
  fi
  echo "done."
else
  echo "Generating output:"
  echo "  Resolution:       $RESOL dpi"
  echo "  Page size:        $PAGESIZE"
  echo "  Output format:    $OUTFORMAT"
  echo "  Output file:      $OUTDIR$OUTFILE"
  case "$OUTFORMAT" in
    txt)
      if [ "$AUTHOR" ] || [ "$TITLE" ] || [ "$SUBJECT" ] || [ "$KEYWORDS" ]; then
        echo
        echo "Warning: Metadata not supported for output format '$OUTFORMAT'." >&2
      fi
      pnmtotiff "$TMPO.pnm" > "$TMPO.tif"
      case ` echo $LANG | awk -F '' '{ print $1 $2 }'` in
        de) TESSLANG="deu" && LANGSET="german";;
        fr) TESSLANG="fra" && LANGSET="france";;
        it) TESSLANG="ita" && LANGSET="italian";;
        nl) TESSLANG="nld" && LANGSET="netherland";;
        pt) TESSLANG="por" && LANGSET="portuguese";;
        es) TESSLANG="spa" && LANGSET="spanish";;
        vi) TESSLANG="vie" && LANGSET="vietnamese";;
        *)  TESSLANG="eng" && LANGSET="english";;
      esac
      echo "Using $LANGSET language setting for OCR..."
      tesseract "$TMPO.tif" "$TMPO" -l "$TESSLANG"
      # You can also use Ocrad or Gocr for OCR
      #ocrad --format=utf8 "$TMPO.pnm" > "$TMPO.txt"
      #gocr -f UTF8 -i "$TMPO.pnm" -o "$TMPO.txt"
      if [ "$OUTFILE_OLD" ]; then
        cp "$OUTFILE_OLD" "$OUTDIR$OUTFILE"
        cat "$TMPO.txt" >> "$OUTDIR$OUTFILE"
        ACTION="(text only) appended"
      else
        mv "$TMPO.txt" "$OUTDIR$OUTFILE"
        ACTION="(text only) written"
      fi
      ;;
    pnm)
      if [ "$AUTHOR" ] || [ "$TITLE" ] || [ "$SUBJECT" ] || [ "$KEYWORDS" ]; then
        echo
        echo "Warning: Metadata not supported for output format '$OUTFORMAT'." >&2
      fi
      mv "$TMPO.pnm" "$OUTDIR$OUTFILE"
      ACTION="written"
      ;;
    tif)
      pnmtotiff "$TMPO.pnm" > "$TMPO.tif"
      tiffset -s  315 "$AUTHOR" -s 269 "$TITLE" -s 270 "$SUBJECT" -s 271 "$KEYWORDS" $TMPO.tif
      if [ "$OUTFILE_OLD" ]; then
        tiffcp "$OUTDIR$OUTFILE_OLD" "$TMPO.tif" "$OUTDIR$OUTFILE"
        ACTION="appended"
      else
        mv "$TMPO.tif" "$OUTDIR$OUTFILE"
        ACTION="written"
      fi
      ;;
    pdf)
      pnmtotiff "$TMPO.pnm" > "$TMPO.tif"
      tiff2pdf -r d -x "$RESOL" -y "$RESOL" -j -p "$PAGESIZE" -o "$TMPO.pdf" "$TMPO.tif"
      if [ "$OUTFILE_OLD" ]; then
        pdftk "$OUTDIR$OUTFILE_OLD" "$TMPO.pdf" cat output "$TMPO.cat.pdf"
        pdfopt "$TMPO.cat.pdf" "$TMPO.opt.pdf"
        ACTION="appended"
      else
        pdfopt "$TMPO.pdf" "$TMPO.opt.pdf"
        ACTION="written"
      fi
      # tiff2pdf does not delete the metadata field even when asked to (e.g. with '-a ""'), so we do not used it to set the metadata. We use pdftk instead.
      # See bug report http://bugzilla.maptools.org/show_bug.cgi?id=2424
      echo -e "InfoBegin\nInfoKey: Author\nInfoValue: $AUTHOR\nInfoBegin\nInfoKey: Title\nInfoValue: $TITLE\nInfoBegin\nInfoKey: Subject\nInfoValue: $SUBJECT\nInfoBegin\nInfoKey: Keywords\nInfoValue: $KEYWORDS" | pdftk "$TMPO.opt.pdf" update_info_utf8 - output "$OUTDIR$OUTFILE"
      ;;
  esac
  echo "done."
  echo "Scanned page $ACTION to '$OUTDIR$OUTFILE'."
fi
echo
